@@ -1,7 +1,10 @@
 /* 
  * Copyright (C) 2006, 2007, 2008 OpenedHand Ltd.
+ * Copyright (C) 2009 Nokia Corporation, all rights reserved.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
+ *         Zeeshan Ali (Khattak) <zeeshanak@gnome.org>
+ *                               <zeeshan.ali@nokia.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -37,6 +40,9 @@
 #include <stdio.h>
 #include <errno.h>
 #include <unistd.h>
+#include <arpa/inet.h>
+#include <net/if.h>
+#include <ifaddrs.h>
 #include <libsoup/soup-headers.h>
 
 #include "gssdp-client.h"
@@ -57,15 +63,24 @@ struct _GSSDPClientPrivate {
         GMainContext      *main_context;
 
         char              *server_id;
+        char              *interface;
+        char              *host_ip;
+
+        GError            **error;
 
         GSSDPSocketSource *request_socket;
         GSSDPSocketSource *multicast_socket;
+
+        gboolean           active;
 };
 
 enum {
         PROP_0,
         PROP_MAIN_CONTEXT,
         PROP_SERVER_ID,
+        PROP_IFACE,
+        PROP_HOST_IP,
+        PROP_ACTIVE,
         PROP_ERROR
 };
 
@@ -86,6 +101,8 @@ static gboolean
 request_socket_source_cb      (gpointer      user_data);
 static gboolean
 multicast_socket_source_cb    (gpointer      user_data);
+static gboolean
+init_network_info             (GSSDPClient  *client);
 
 static void
 gssdp_client_init (GSSDPClient *client)
@@ -95,12 +112,25 @@ gssdp_client_init (GSSDPClient *client)
                                          GSSDP_TYPE_CLIENT,
                                          GSSDPClientPrivate);
 
+        client->priv->active = TRUE;
+
         /* Generate default server ID */
         client->priv->server_id = make_server_id ();
+}
+
+static void
+gssdp_client_constructed (GObject *object)
+{
+        GSSDPClient *client = GSSDP_CLIENT (object);
+
+        /* Make sure all network info is available to us */
+        if (!init_network_info (client))
+                return;
 
         /* Set up sockets (Will set errno if it failed) */
         client->priv->request_socket =
-                gssdp_socket_source_new (GSSPP_SOCKET_SOURCE_TYPE_REQUEST);
+                gssdp_socket_source_new (GSSPP_SOCKET_SOURCE_TYPE_REQUEST,
+                                         gssdp_client_get_host_ip (client));
         if (client->priv->request_socket != NULL) {
                 g_source_set_callback
                         ((GSource *) client->priv->request_socket,
@@ -110,7 +140,8 @@ gssdp_client_init (GSSDPClient *client)
         }
 
         client->priv->multicast_socket =
-                gssdp_socket_source_new (GSSDP_SOCKET_SOURCE_TYPE_MULTICAST);
+                gssdp_socket_source_new (GSSDP_SOCKET_SOURCE_TYPE_MULTICAST,
+                                         gssdp_client_get_host_ip (client));
         if (client->priv->multicast_socket != NULL) {
                 g_source_set_callback
                         ((GSource *) client->priv->multicast_socket,
@@ -118,6 +149,24 @@ gssdp_client_init (GSSDPClient *client)
                          client,
                          NULL);
         }
+
+        if (!client->priv->request_socket || !client->priv->multicast_socket) {
+                if (client->priv->error)
+                        g_set_error_literal (client->priv->error,
+                                             GSSDP_ERROR,
+                                             GSSDP_ERROR_FAILED,
+                                             strerror (errno));
+
+                return;
+        }
+
+        g_source_attach ((GSource *) client->priv->request_socket,
+                         client->priv->main_context);
+        g_source_unref ((GSource *) client->priv->request_socket);
+
+        g_source_attach ((GSource *) client->priv->multicast_socket,
+                         client->priv->main_context);
+        g_source_unref ((GSource *) client->priv->multicast_socket);
 }
 
 static void
@@ -142,6 +191,17 @@ gssdp_client_get_property (GObject    *object,
                          (gpointer)
                           gssdp_client_get_main_context (client));
                 break;
+        case PROP_IFACE:
+                g_value_set_string (value,
+                                    gssdp_client_get_interface (client));
+                break;
+        case PROP_HOST_IP:
+                g_value_set_string (value,
+                                    gssdp_client_get_host_ip (client));
+                break;
+        case PROP_ACTIVE:
+                g_value_set_boolean (value, client->priv->active);
+                break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
                 break;
@@ -168,18 +228,13 @@ gssdp_client_set_property (GObject      *object,
                                                g_value_get_pointer (value));
                 break;
         case PROP_ERROR:
-                if (!client->priv->request_socket ||
-                    !client->priv->multicast_socket) {
-                        GError **error;
-
-                        error = g_value_get_pointer (value);
-
-                        g_set_error (error,
-                                     GSSDP_ERROR,
-                                     GSSDP_ERROR_FAILED,
-                                     strerror (errno));
-                }
-
+                client->priv->error = g_value_get_pointer (value);
+                break;
+        case PROP_IFACE:
+                client->priv->interface = g_value_dup_string (value);
+                break;
+        case PROP_ACTIVE:
+                client->priv->active = g_value_get_boolean (value);
                 break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -220,6 +275,8 @@ gssdp_client_finalize (GObject *object)
         client = GSSDP_CLIENT (object);
 
         g_free (client->priv->server_id);
+        g_free (client->priv->interface);
+        g_free (client->priv->host_ip);
 }
 
 static void
@@ -229,6 +286,7 @@ gssdp_client_class_init (GSSDPClientClass *klass)
 
 	object_class = G_OBJECT_CLASS (klass);
 
+	object_class->constructed = gssdp_client_constructed;
 	object_class->set_property = gssdp_client_set_property;
 	object_class->get_property = gssdp_client_get_property;
 	object_class->dispose      = gssdp_client_dispose;
@@ -288,6 +346,65 @@ gssdp_client_class_init (GSSDPClientClass *klass)
                           G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                           G_PARAM_STATIC_BLURB));
 
+        /**
+         * GSSDPClient:interface
+         *
+         * The name of the network interface this client is associated with.
+         * Set to NULL to autodetect.
+         **/
+        g_object_class_install_property
+                (object_class,
+                 PROP_IFACE,
+                 g_param_spec_string
+                         ("interface",
+                          "Network interface",
+                          "The name of the associated network interface.",
+                          NULL,
+                          G_PARAM_READWRITE |
+                          G_PARAM_CONSTRUCT_ONLY |
+                          G_PARAM_STATIC_NAME |
+                          G_PARAM_STATIC_NICK |
+                          G_PARAM_STATIC_BLURB));
+
+        /**
+         * GSSDPClient:host-ip
+         *
+         * The IP address of the assoicated network interface.
+         **/
+        g_object_class_install_property
+                (object_class,
+                 PROP_HOST_IP,
+                 g_param_spec_string ("host-ip",
+                                      "Host IP",
+                                      "The IP address of the associated"
+                                      "network interface",
+                                      NULL,
+                                      G_PARAM_READABLE |
+                                      G_PARAM_STATIC_NAME |
+                                      G_PARAM_STATIC_NICK |
+                                      G_PARAM_STATIC_BLURB));
+
+        /**
+         * GSSDPClient:active
+         *
+         * Whether this client is active or not (passive). When active
+         * (default), the client sends messages on the network, otherwise
+         * not. In most cases, you don't want to touch this property.
+         *
+         **/
+        g_object_class_install_property
+                (object_class,
+                 PROP_ACTIVE,
+                 g_param_spec_boolean
+                         ("active",
+                          "Active",
+                          "TRUE if the client is active.",
+                          TRUE,
+                          G_PARAM_READWRITE |
+                          G_PARAM_STATIC_NAME |
+                          G_PARAM_STATIC_NICK |
+                          G_PARAM_STATIC_BLURB));
+
         /**
          * GSSDPClient::message-received
          *
@@ -312,17 +429,20 @@ gssdp_client_class_init (GSSDPClientClass *klass)
 
 /**
  * gssdp_client_new
- * @main_context: The #GMainContext to associate with
+ * @main_context: The #GMainContext to associate with, or NULL
+ * @interface: The name of the network interface, or %NULL for auto-detection.
  * @error: Location to store error, or NULL
  *
  * Return value: A new #GSSDPClient object.
  **/
 GSSDPClient *
 gssdp_client_new (GMainContext *main_context,
+                  const char   *interface,
                   GError      **error)
 {
         return g_object_new (GSSDP_TYPE_CLIENT,
                              "main-context", main_context,
+                             "interface", interface,
                              "error", error,
                              NULL);
 }
@@ -336,23 +456,10 @@ gssdp_client_set_main_context (GSSDPClient  *client,
 {
         g_return_if_fail (GSSDP_IS_CLIENT (client));
 
+        /* A NULL main_context is fine */
         if (main_context)
                 client->priv->main_context = g_main_context_ref (main_context);
 
-        /* A NULL main_context is fine */
-        
-        if (client->priv->request_socket) {
-                g_source_attach ((GSource *) client->priv->request_socket,
-                                 client->priv->main_context);
-                g_source_unref ((GSource *) client->priv->request_socket);
-        }
-
-        if (client->priv->multicast_socket) {
-                g_source_attach ((GSource *) client->priv->multicast_socket,
-                                 client->priv->main_context);
-                g_source_unref ((GSource *) client->priv->multicast_socket);
-        }
-
         g_object_notify (G_OBJECT (client), "main-context");
 }
 
@@ -360,7 +467,7 @@ gssdp_client_set_main_context (GSSDPClient  *client,
  * gssdp_client_get_main_context
  * @client: A #GSSDPClient
  *
- * Return value: The #GMainContext @client is associated with.
+ * Return value: The #GMainContext @client is associated with, or NULL.
  **/
 GMainContext *
 gssdp_client_get_main_context (GSSDPClient *client)
@@ -408,6 +515,52 @@ gssdp_client_get_server_id (GSSDPClient *client)
         return client->priv->server_id;
 }
 
+/**
+ * gssdp_client_get_interface
+ * @client: A #GSSDPClient
+ *
+ * Get the name of the network interface associated to @client.
+ *
+ * Return value: The network interface name. This string should not be freed.
+ **/
+const char *
+gssdp_client_get_interface (GSSDPClient *client)
+{
+        g_return_val_if_fail (GSSDP_IS_CLIENT (client), NULL);
+
+        return client->priv->interface;
+}
+
+/**
+ * gssdp_client_get_host_ip
+ * @client: A #GSSDPClient
+ *
+ * Get the IP address we advertise ourselves as using.
+ *
+ * Return value: The IP address. This string should not be freed.
+ **/
+const char *
+gssdp_client_get_host_ip (GSSDPClient *client)
+{
+        g_return_val_if_fail (GSSDP_IS_CLIENT (client), NULL);
+
+        return client->priv->host_ip;
+}
+
+/**
+ * gssdp_client_get_active
+ * @client: A #GSSDPClient
+ *
+ * Return value: %TRUE if @client is active, %FALSE otherwise.
+ **/
+gboolean
+gssdp_client_get_active (GSSDPClient *client)
+{
+        g_return_val_if_fail (GSSDP_IS_CLIENT (client), FALSE);
+
+        return client->priv->active;
+}
+
 /**
  * _gssdp_client_send_message
  * @client: A #GSSDPClient
@@ -429,6 +582,10 @@ _gssdp_client_send_message (GSSDPClient *client,
         g_return_if_fail (GSSDP_IS_CLIENT (client));
         g_return_if_fail (message != NULL);
 
+        if (!client->priv->active)
+                /* We don't send messages in passive mode */
+                return;
+
         /* Broadcast if @dest_ip is NULL */
         if (dest_ip == NULL)
                 dest_ip = SSDP_ADDR;
@@ -554,6 +711,9 @@ socket_source_cb (GSSDPSocketSource *socket, GSSDPClient *client)
         struct sockaddr_in addr;
         socklen_t addr_size;
         SoupMessageHeaders *headers;
+        struct in_addr our_addr;
+        in_addr_t our_network;
+        in_addr_t recv_network;
 
         /* Get FD */
         fd = gssdp_socket_source_get_fd (socket);
@@ -568,6 +728,18 @@ socket_source_cb (GSSDPSocketSource *socket, GSSDPClient *client)
                           (struct sockaddr *) &addr,
                           &addr_size);
 
+        /* We need the following lines to make sure the right client received
+         * the packet. We won't need to do this if there was any way to tell
+         * Mr. Unix that we are only interested in receiving multicast packets
+         * on this socket from a particular interface but AFAIK that is not
+         * possible, at least not in a portable way.
+         */
+        recv_network = inet_netof (addr.sin_addr);
+        our_addr.s_addr = inet_addr (gssdp_client_get_host_ip (client));
+        our_network = inet_netof (our_addr);
+        if (recv_network != our_network)
+                return TRUE;
+
         if (bytes >= BUF_SIZE) {
                 g_warning ("Received packet of %u bytes, but the maximum "
                            "buffer size is %d. Packed dropped.",
@@ -642,3 +814,114 @@ multicast_socket_source_cb (gpointer user_data)
 
         return socket_source_cb (client->priv->multicast_socket, client);
 }
+
+/*
+ * Get the host IP for the specified interface. If no interface is specified,
+ * it gets the IP of the first up & running interface and sets @interface
+ * appropriately.
+ */
+static char *
+get_host_ip (char **interface)
+{
+        struct ifaddrs *ifa_list, *ifa;
+        char *ret;
+        GList *up_ifaces, *iface;
+
+        ret = NULL;
+        up_ifaces = NULL;
+
+        if (getifaddrs (&ifa_list) != 0) {
+                g_error ("Failed to retrieve list of network interfaces:\n%s\n",
+                         strerror (errno));
+
+                return NULL;
+        }
+
+        for (ifa = ifa_list; ifa != NULL; ifa = ifa->ifa_next) {
+                if (ifa->ifa_addr == NULL)
+                        continue;
+
+                if (*interface && strcmp (*interface, ifa->ifa_name) != 0)
+                        continue;
+                else if (!(ifa->ifa_flags & IFF_UP))
+                        continue;
+
+                /* Loopback and IPv6 interfaces go at the bottom on the list */
+                if (ifa->ifa_flags & IFF_LOOPBACK ||
+                    ifa->ifa_addr->sa_family == AF_INET6)
+                        up_ifaces = g_list_append (up_ifaces, ifa);
+                else
+                        up_ifaces = g_list_prepend (up_ifaces, ifa);
+        }
+
+        for (iface = up_ifaces; iface != NULL; iface = iface->next) {
+                char ip[INET6_ADDRSTRLEN];
+                const char *p;
+                struct sockaddr_in *s4;
+                struct sockaddr_in6 *s6;
+
+                p = NULL;
+
+                ifa = iface->data;
+
+                switch (ifa->ifa_addr->sa_family) {
+                case AF_INET:
+                        s4 = (struct sockaddr_in *) ifa->ifa_addr;
+                        p = inet_ntop (AF_INET,
+                                       &s4->sin_addr, ip, sizeof (ip));
+                        break;
+                case AF_INET6:
+                        s6 = (struct sockaddr_in6 *) ifa->ifa_addr;
+                        p = inet_ntop (AF_INET6,
+                                       &s6->sin6_addr, ip, sizeof (ip));
+                        break;
+                default:
+                        continue; /* Unknown: ignore */
+                }
+
+                if (p != NULL) {
+                        ret = g_strdup (p);
+
+                        if (*interface == NULL)
+                                *interface = g_strdup (ifa->ifa_name);
+                        break;
+                }
+        }
+
+        g_list_free (up_ifaces);
+        freeifaddrs (ifa_list);
+
+        return ret;
+}
+
+static gboolean
+init_network_info (GSSDPClient *client)
+{
+        gboolean ret = TRUE;
+
+        if (client->priv->interface == NULL || client->priv->host_ip == NULL)
+                client->priv->host_ip =
+                        get_host_ip (&client->priv->interface);
+
+        if (client->priv->interface == NULL) {
+                if (client->priv->error)
+                        g_set_error (client->priv->error,
+                                     GSSDP_ERROR,
+                                     GSSDP_ERROR_FAILED,
+                                     "No default route?");
+
+                ret = FALSE;
+        } else if (client->priv->host_ip == NULL) {
+                if (client->priv->error)
+                        g_set_error (client->priv->error,
+                                     GSSDP_ERROR,
+                                     GSSDP_ERROR_FAILED,
+                                     "Failed to find IP of interface %s",
+                                     client->priv->interface);
+
+                ret = FALSE;
+        }
+
+        return ret;
+}
+
